{***********************************************************}
{*                                                         *}
{* ICS 4410                                                *}
{* Fall '87                                                *}
{*                                                         *}
{* Lynn R. Akers, Jr.                                      *}
{*                                                         *}
{* This program has been modified to run on an IBM PC or   *}
{* compatable using TURBO Pascal version 3.0.              *}
{*                                                         *}
{* It has only been tested for MICRO, so some bugs may     *}
{* still exist.                                            *}
{*                                                         *}
{*   Joseph Bergin                                         *}
{*   Modified for the Macintosh, TML Pascal   1989         *}
{*                                                         *}
{*                                                         *}
{***********************************************************}

PROGRAM sam2 (input, outFile, memf, InFile);
{version of the SAM assembler for the MACC2 byte-addressable machine}
{        Author: C.T. Wilkes (from original by A.B. Maccabe)        }
{                   Last modified: 84/11/26 CTW                     }

	CONST
		maxSmallInt = 16383;
		minSmallInt = -16383;
	TYPE
		word = SET OF 0..15;

		bytee = SET OF 0..7;

		wordrange = minSmallInt..maxSmallInt;

		byterange = -128..127;

		opkind = (OIN, IA, IS, IM, IDENT, FN, FA, FS, FM, FD, BI, BO, BA, IC, FC, JSR, BKT, LD, STO, LDA, FLT, FIX, J, SR, SL, RD, WR, TRNG, HALT, NOP, CLR, REALDIR, STRINGDIR, INTDIR, SKIPDIR, LABELDIR);

		errorkind = (notErr, unkopname, badregaddr, badgenaddr, badint, badreal, badstr, badname, illregaddr, illimedaddr, badshftamt, badstring, invalidname);
		warnkind = (longname, missingR, missingnum, missingrp, missingcomma, namedefined, badskip, esctooshort, strtooshort, textfollows);

		icrec = RECORD
				CASE boolean OF
					true: (
							wdf: word
					);
					false: (
							sintf: wordrange
					)
			END;

		bcrec = RECORD
				CASE boolean OF
					true: (
							bytef: bytee
					);
					false: (
							hintf: byterange
					)
			END;

		stringtype = PACKED ARRAY[1..25] OF char;



		name = PACKED ARRAY[1..7] OF char;

		symptr = ^symrec;
		symrec = RECORD
				id: name;
				left, right: symptr;
				patch, loc: integer
			END;

	VAR
		Inop, Iaop, Isop, Imop, Idop, Fnop, Faop, Fsop, Fmop, Fdop, Biop, Boop, Baop, Icop, Fcop, Jsrop, Bktop, Ldop, Stoop, Ldaop, Fltop, Fixop, Jop, Srop, Slop, Rdop, Wrop, Trngop, Haltop: word;

		InFile, OutFile: text;

		Saved: boolean;  {flag for one character lookahead}
		Ch: char;   {Current character from the input}

		Mem: ARRAY[0..maxSmallInt] OF bytee; {Memory Image being created}
		Lc: 0..maxSmallInt;  {Location Counter}
		memf: FILE OF byterange;    {File for the memory image}

		Symbols: symptr;
		cLine: integer;  {Number of the current input line}
		Listing: boolean;

		Error: errorkind;
		Errs: boolean;
		Warning: boolean;
		Morewarn: boolean;
		Warns: ARRAY[1..10] OF warnkind;
		Windex: integer;

		Icr: icrec;{Integer/Word Conversion Record}
		Bcr: bcrec;{Halfint/Byte Conversion Record}
		i: integer;{loop index}

		Arg, Memfname, Srcfname: stringtype;
		CodeFileName, ListingFileName: STRING;
		VRefNum: integer;
		finderInfo: FInfo;
		noError: OSErr;

	PROCEDURE clearLine (n: integer);
		VAR
			r: rect;
	BEGIN
	r.top := n - 15;
	r.left := 1;
	r.bottom := n + 15;
	r.right := 396;
	eraseRect(r);
	END;

	FUNCTION OldFileName (p: STRING): STRING;
		VAR
			topLeft: point;
			TypeList: SFTypeList;
			reply: SFReply;
	BEGIN
	topleft.v := 100;
	topLeft.h := 100;
	typelist[0] := 'TEXT';
	OldFileName := '';
	sfGetFile(topLeft, '', NIL, 1, typeList, NIL, reply);
	IF reply.good THEN BEGIN
		OldFileName := (reply.fName);
		vRefNum := reply.vRefNum;
		END;
	END;

	FUNCTION NewFileName (p: STRING): STRING;
		VAR
			topLeft: point;
			reply: SFReply;
			NFN: STRING;
	BEGIN
	topleft.v := 100;
	topLeft.h := 100;
	NFN := 'Listing.T';
	NewFileName := NFN;
	sfPutFile(topLeft, '', NFN, NIL, reply);
	IF reply.good THEN
		NewFileName := (reply.fName);
	END;



	PROCEDURE wordtobytes (w: word; VAR hibytee, lobytee: bytee);
		VAR
			i: integer;
	BEGIN
	lobytee := [];
	hibytee := [];
	FOR i := 0 TO 7 DO BEGIN
		IF i IN w THEN
			lobytee := lobytee + [i];
		IF i + 8 IN w THEN
			hibytee := hibytee + [i]
		END {for}
	END; {procedure wordtobytes}


	PROCEDURE bytestoword (hibytee, lobytee: bytee; VAR w: word);
		VAR
			i: integer;
	BEGIN
	w := [];
	FOR i := 0 TO 7 DO BEGIN
		IF i IN lobytee THEN
			w := w + [i];
		IF i IN hibytee THEN
			w := w + [i + 8]
		END {for}
	END; {procedure bytestoword}


	PROCEDURE insertmem (w: word);
		VAR
			lobytee, hibytee: bytee;
	BEGIN
	wordtobytes(w, hibytee, lobytee);
	Mem[Lc] := hibytee;
	Mem[Lc + 1] := lobytee;
	Lc := Lc + 2
	END; {procedure insertmem}

	PROCEDURE checktab (cur: symptr);
	BEGIN
	IF cur <> NIL THEN BEGIN
		checktab(cur^.left);
		IF cur^.loc < 0 THEN BEGIN
			Warning := true;
			writeln(outFile, 'WARNING -- ', cur^.id, ' Undefined');
			END;
		checktab(cur^.right)
		END {if}
	END; {procedure checktab}


	PROCEDURE warn (w: warnkind);
	BEGIN
	IF NOT Morewarn THEN
		Windex := 1;
	Morewarn := true;
	Warns[Windex] := w;
	Windex := Windex + 1
	END; {procedure warn}


	PROCEDURE printwarn;
		VAR
			i: integer;
	BEGIN
	write(outFile, 'WARNING -- ');
	CASE Warns[1] OF
	textfollows: 
		write(outFile, 'Text follows instruction');
	esctooshort: 
		write(outFile, 'Need 3 digits to specify an unprintable character');
	strtooshort: 
		write(outFile, 'Missing " in string');
	badskip: 
		write(outFile, 'Skip value must be positive, Skip directive ignored');
	namedefined: 
		write(outFile, 'Name already defined, earlier definition lost');
	longname: 
		write(outFile, 'Name too long, only 7 characters used');
	missingR: 
		write(outFile, 'Missing R in Register Address');
	missingnum: 
		write(outFile, 'Missing Number in Register Address (0 assumed)');
	missingrp: 
		write(outFile, 'Missing ")" in Indexed Address');
	missingcomma: 
		write(outFile, 'Missing ","')
	END; {case}
	writeln(outFile, ' on line ', cLine : 1);

	FOR i := 2 TO Windex - 1 DO
		Warns[i - 1] := Warns[i];
	Windex := Windex - 1;
	IF Windex <= 1 THEN
		Morewarn := false
	END; {procedure printwarn}


	PROCEDURE inregaddr (VAR w: word; reg: integer; hbit: integer);
    {Insert a register address into a word}
	BEGIN
	IF reg > 7 THEN BEGIN
		w := w + [hbit];
		reg := reg - 8
		END; {if reg > 7}
	hbit := hbit - 1;
	IF reg > 3 THEN BEGIN
		w := w + [hbit];
		reg := reg - 4
		END; {if reg > 3}
	hbit := hbit - 1;
	IF reg > 1 THEN BEGIN
		w := w + [hbit];
		reg := reg - 2
		END; {if reg > 1}
	hbit := hbit - 1;
	IF reg > 0 THEN
		w := w + [hbit]
	END; {procedure inregaddr}


	PROCEDURE getch;
    {Get a character from the input -- character may have been saved}
	BEGIN
	IF NOT Saved THEN
		IF eof(InFile) THEN
			Ch := '%'
		ELSE IF NOT eoln(InFile) THEN BEGIN
			REPEAT
				read(InFile, Ch);
				IF Listing THEN
					write(outFile, Ch)
			UNTIL (Ch <> ' ') AND (Ch <> chr(9)) OR eoln(InFile);
			IF Ch = '%' THEN BEGIN
				WHILE NOT eoln(InFile) DO BEGIN
					read(InFile, Ch);
					IF Listing THEN
						write(outFile, Ch)
					END; {while}
				Ch := '%'
				END
			END
		ELSE
			Ch := '%'
	ELSE
		Saved := false
	END; {procedure getch}


	PROCEDURE scanname (VAR id: name);
		VAR
			i: integer;
	BEGIN
	FOR i := 1 TO 7 DO
		id[i] := ' ';
	i := 1;
	WHILE (Ch IN ['a'..'z', 'A'..'Z', '0'..'9', '$']) AND (i <= 7) DO BEGIN
		id[i] := Ch;
		i := i + 1;
		getch
		END; {while}
	IF Ch IN ['a'..'z', 'A'..'Z', '0'..'9', '$'] THEN
		warn(longname);
	WHILE Ch IN ['a'..'z', 'A'..'Z', '0'..'9', '$'] DO
		getch;
	Saved := true
	END; {procedure scanname}


	FUNCTION findname (id: name): symptr;
		VAR
			temp: symptr;
			found: boolean;
	BEGIN
	temp := Symbols;
	found := false;
	WHILE NOT found AND (temp <> NIL) DO
		IF temp^.id = id THEN
			found := true
		ELSE IF temp^.id > id THEN
			temp := temp^.left
		ELSE
			temp := temp^.right;
	findname := temp
	END; {function findname}


	FUNCTION inname (id: name): symptr;
		VAR
			cur, prev: symptr;
	BEGIN
	cur := Symbols;
	prev := NIL;
	WHILE cur <> NIL DO BEGIN
		prev := cur;
		IF cur^.id > id THEN
			cur := cur^.left
		ELSE
			cur := cur^.right
		END; {while}
	new(cur);
	cur^.left := NIL;
	cur^.right := NIL;
	cur^.id := id;
	IF prev = NIL THEN
		Symbols := cur
	ELSE IF prev^.id > id THEN
		prev^.left := cur
	ELSE
		prev^.right := cur;
	inname := cur
	END; {function inname}


	PROCEDURE scanstr;
		TYPE
			ccrec = RECORD
					CASE boolean OF
						true: (
								wf: word
						);
						false: (
								c: PACKED ARRAY[1..2] OF char;
						) {Inverted Bytes!!!}
				END;
		VAR
			one: boolean;
			ccr: ccrec;
			ival: integer;
	BEGIN
	one := true;
	getch;
	IF Ch = '"' THEN BEGIN
		IF NOT eoln(InFile) THEN BEGIN
			read(InFile, Ch);
			IF Listing THEN
				write(outFile, Ch)
			END; {if}
		WHILE (Ch <> '"') AND NOT eoln(InFile) DO BEGIN
			IF Ch = ':' THEN BEGIN
				IF NOT eoln(InFile) THEN BEGIN
					read(InFile, Ch);
					IF Listing THEN
						write(outFile, Ch);
					IF Ch IN ['0'..'9'] THEN BEGIN
						ival := ord(Ch) - ord('0');
						IF NOT eoln(InFile) THEN BEGIN
							read(InFile, Ch);
							IF Listing THEN
								write(outFile, Ch);
							IF Ch IN ['0'..'9'] THEN BEGIN
								ival := ival * 10 + ord(Ch) - ord('0');
								IF NOT eoln(InFile) THEN BEGIN
									read(InFile, Ch);
									IF Listing THEN
										write(outFile, Ch);
									IF Ch IN ['0'..'9'] THEN
										ival := ival * 10 + ord(Ch) - ord('0')
									ELSE
										warn(esctooshort)
									END
								ELSE
									warn(esctooshort)
								END
							ELSE
								warn(esctooshort)
							END
						ELSE
							warn(esctooshort);
						Ch := chr(ival)
						END
					END
				ELSE
					warn(esctooshort)
				END;
			IF one THEN BEGIN
				one := false;
				ccr.c[1] := Ch
				END
			ELSE BEGIN
				one := true;
				ccr.c[2] := Ch;
				insertmem(ccr.wf)
				END; {else}
			IF NOT eoln(InFile) THEN BEGIN
				read(InFile, Ch);
				IF Listing THEN
					write(outFile, Ch)
				END {if}
			END; {while}
		IF one THEN BEGIN
			ccr.c[1] := chr(0);
			ccr.c[2] := chr(0);
			END
		ELSE
			ccr.c[2] := chr(0);
		insertmem(ccr.wf);
		IF Ch <> '"' THEN
			warn(strtooshort)
		END
	ELSE
		Error := badstring
	END; {procedure scanstr}


	PROCEDURE scanreal (VAR w1, w2: word);
		TYPE
			bytetype = SET OF 0..7;

			mrec = RECORD
					byte1: bytee;
{    byte2: bytee;}
{    byte3: bytee;}
					byte4: bytee;
					byte5: bytee;
					byte6: bytee
				END;

			rcrec = RECORD
					CASE boolean OF
						true: (
								rf: real
						);
						false: (
								mf: mrec
						)
				END;

		VAR
			rcr: rcrec;
			dval: real;

	BEGIN
	rcr.rf := 0.0;
	getch;
	WHILE Ch IN ['0'..'9'] DO BEGIN
		rcr.rf := rcr.rf * 10 + ord(Ch) - ord('0');
		getch
		END; {while}
	IF Ch = '.' THEN BEGIN
		getch;
		dval := 10.0;
		WHILE Ch IN ['0'..'9'] DO BEGIN
			rcr.rf := rcr.rf + (ord(Ch) - ord('0')) / dval;
			dval := dval * 10.0;
			getch
			END {while}
		END
	ELSE
		Saved := true;
	bytestoword(rcr.mf.byte1, rcr.mf.byte4, w1);
	bytestoword(rcr.mf.byte5, rcr.mf.byte6, w2)
	END; {procedure scanreal}


	PROCEDURE scanint (VAR w: word);
    {Get an integer from the input stream}
		VAR
			temp: integer;
			neg: boolean;
	BEGIN
	neg := false;
	temp := 0;
	getch;
	IF Ch IN ['-', '+'] THEN BEGIN
		IF Ch = '-' THEN
			neg := true;
		getch
		END; {if Ch in ['-', '+']}
	WHILE Ch IN ['0'..'9'] DO BEGIN
		temp := temp * 10 + ord(Ch) - ord('0');
		getch
		END; {while}
	Saved := true;    {Note the lookahead}
	IF neg THEN
		temp := -temp;
	IF (temp > maxSmallInt) OR (temp < -maxSmallInt) THEN
		Error := badint
	ELSE BEGIN
		Icr.sintf := temp;
		w := Icr.wdf
		END
	END; { scanint }


	FUNCTION getregaddr: integer;
    {Get a register address from the input stream}
		VAR
			temp: integer;

	BEGIN
	getch;
	IF Ch = 'R' THEN
		getch
	ELSE
		warn(missingR);
	IF Ch IN ['0'..'9'] THEN BEGIN  {Check first digit}
		temp := ord(Ch) - ord('0');
		getch;
		IF Ch IN ['0'..'9'] THEN 		{Check for two digits}
			temp := temp * 10 + ord(Ch) - ord('0')
		ELSE
			Saved := true;
		IF temp > 15 THEN
			Error := badregaddr
		END
	ELSE  {if Ch is a digit}
		warn(missingnum);

	getregaddr := temp
	END; {function getregaddr}


	PROCEDURE getgenaddr (op: opkind; VAR w1, w2: word; VAR flag: boolean);
		VAR
			reg: integer;
			id: name;
			idrec: symptr;

	BEGIN
	flag := false;
	getch;
	IF Ch = '*' THEN BEGIN
		w1 := w1 + [6];
		getch
		END; {if Ch = '*'}

	IF Ch IN ['a', 'z', 'A'..'Q', 'S'..'Z', '$'] THEN BEGIN
		flag := true;
		scanname(id);
		idrec := findname(id);
		IF idrec = NIL THEN BEGIN
			idrec := inname(id);
			idrec^.loc := -1;
			idrec^.patch := Lc + 2;
			Icr.sintf := -1
			END
		ELSE IF idrec^.loc = -1 THEN BEGIN
			Icr.sintf := idrec^.patch;
			idrec^.patch := Lc + 2
			END
		ELSE
			Icr.sintf := idrec^.loc;
		w2 := Icr.wdf;
		getch;
		IF Ch = '(' THEN BEGIN
			w1 := w1 + [5];
			reg := getregaddr;
			IF Error = notErr THEN BEGIN
				inregaddr(w1, reg, 3);
				getch;
				IF Ch <> ')' THEN
					warn(missingrp)
				END {if Error in register address}
			END
		ELSE     {Ch <> '('}
			w1 := w1 + [4]
		END (* Letter other than 'R' *)
	ELSE IF Ch IN ['0'..'9'] THEN BEGIN
		Saved := true;
		w1 := w1 + [4];
		flag := true;
		scanint(w2)
		END
	ELSE IF Ch IN ['R', '#', '-', '+', '&'] THEN
		CASE Ch OF
		'R':  BEGIN   {direct register}
			flag := false;
			Saved := true;
			reg := getregaddr;
			inregaddr(w1, reg, 3);
			IF (op IN [JSR, BKT, LDA, J]) AND NOT (6 IN w1) THEN
				Error := illregaddr
			END;
		'#':  BEGIN   {immediate}
			flag := true;
			IF 6 IN w1 THEN
				Error := badgenaddr
			ELSE IF op IN [FN, FA, FS, FM, FD, FC, FIX, JSR, BKT, STO, J, RD, TRNG] THEN
				Error := illimedaddr
			ELSE IF w1 = Wrop + [7] THEN
				Error := illimedaddr
			ELSE IF w1 = Wrop + [10, 7] THEN
				Error := illimedaddr
			ELSE BEGIN
				w1 := w1 + [4, 5];
				scanint(w2)
				END {else}
			END;
		'-', '+':  BEGIN   {indexed}
			w1 := w1 + [5];
			flag := true;
			IF Ch = '-' THEN
				Saved := true;
			scanint(w2);
			getch;
			IF Ch = '(' THEN BEGIN
				reg := getregaddr;
				IF Error = notErr THEN BEGIN
					inregaddr(w1, reg, 3);
					getch;
					IF Ch <> ')' THEN
						warn(missingrp)
					END {if Error in register address}
				END
			ELSE   {Ch <> '('}
				Error := badgenaddr
			END;
		'&':  BEGIN
			flag := true;
			IF 6 IN w1 THEN
				Error := badgenaddr
			ELSE BEGIN
				w1 := w1 + [4, 5, 6];
				scanint(w2)
				END
			END
		END {case of Ch}
	ELSE
		Error := badgenaddr
	END; {procedure getgenaddr}

	PROCEDURE getBop (VAR op: opkind; VAR wd: word);
    {Get an operator or directive name that begins with 'B'}
	BEGIN
	getch;
	IF Ch IN ['A', 'I', 'K', 'O'] THEN
		CASE Ch OF
		'A':  BEGIN
			op := BA;
			wd := Baop
			END;
		'I':  BEGIN
			op := BI;
			wd := Biop
			END;
		'K':  BEGIN
			getch;
			IF Ch = 'T' THEN BEGIN
				op := BKT;
				wd := Bktop
				END
			ELSE   {Ch <> 'T'}
				Error := unkopname
			END;
		'O':  BEGIN
			op := BO;
			wd := Boop
			END
		END {case of Ch}
	ELSE      {character does not legally follow `B'}
		Error := unkopname
	END; {procedure getBop}


	PROCEDURE getFop (VAR op: opkind; VAR wd: word);
    {Get an operator or directive name that begins with 'F'}
	BEGIN
	getch;
	IF Ch IN ['A', 'C', 'D', 'I', 'L', 'M', 'N', 'S'] THEN
		CASE Ch OF
		'A':  BEGIN
			op := FA;
			wd := Faop
			END;
		'C':  BEGIN
			op := FC;
			wd := Fcop
			END;
		'D':  BEGIN
			op := FD;
			wd := Fdop
			END;
		'I':  BEGIN
			getch;
			IF Ch = 'X' THEN BEGIN
				op := FIX;
				wd := Fixop
				END
			ELSE 		{Ch <> 'X'}
				Error := unkopname
			END;
		'L':  BEGIN
			getch;
			IF Ch = 'T' THEN BEGIN
				op := FLT;
				wd := Fltop
				END
			ELSE 		{Ch <> 'T'}
				Error := unkopname
			END;
		'M':  BEGIN
			op := FM;
			wd := Fmop
			END;
		'N':  BEGIN
			op := FN;
			wd := Fnop
			END;
		'S':  BEGIN
			op := FS;
			wd := Fsop
			END
		END {case of Ch}
	ELSE 					{character does not legally follow `F'}
		Error := unkopname
	END; {procedure getFop}


	PROCEDURE getIop (VAR op: opkind; VAR wd: word);
    {Get an operator or directive name that begins with 'I'}
	BEGIN
	getch;
	IF Ch IN ['A', 'C', 'D', 'M', 'N', 'S'] THEN
		CASE Ch OF
		'A':  BEGIN
			op := IA;
			wd := Iaop
			END;
		'C':  BEGIN
			op := IC;
			wd := Icop
			END;
		'D':  BEGIN
			op := IDENT;
			wd := Idop
			END;
		'M':  BEGIN
			op := IM;
			wd := Imop
			END;
		'N':  BEGIN
			getch;
			IF Ch = 'T' THEN
				op := INTDIR
			ELSE BEGIN
				op := OIN;
				wd := Inop;
				Saved := true
				END
			END;
		'S':  BEGIN
			op := IS;
			wd := Isop
			END
		END {case of Ch}
	ELSE 					{character does not legally follow `I'}
		Error := unkopname
	END; {getIop}


	PROCEDURE getJop (VAR op: opkind; VAR wd: word);
    {Get an operator or directive name that begins with 'J'}
	BEGIN
	getch;
	IF Ch IN ['E', 'G', 'L', 'M', 'N', 'S'] THEN BEGIN
		op := J; {most are simple jumps--except JSR!!}
		CASE Ch OF
		'E':  BEGIN
			getch;
			IF Ch = 'Q' THEN
				wd := Jop + [7, 8]
			ELSE
				Error := unkopname
			END;
		'G':  BEGIN
			getch;
			IF Ch = 'E' THEN
				wd := Jop + [7, 9]
			ELSE IF Ch = 'T' THEN
				wd := Jop + [8, 9]
			ELSE
				Error := unkopname
			END;
		'L':  BEGIN
			getch;
			IF Ch = 'E' THEN
				wd := Jop + [8]
			ELSE IF Ch = 'T' THEN
				wd := Jop + [7]
			ELSE
				Error := unkopname
			END;
		'M':  BEGIN
			getch;
			IF Ch = 'P' THEN
				wd := Jop
			ELSE
				Error := unkopname
			END;
		'N':  BEGIN
			getch;
			IF Ch = 'E' THEN
				wd := Jop + [9]
			ELSE
				Error := unkopname
			END;
		'S':  BEGIN
			getch;
			IF Ch = 'R' THEN BEGIN
				op := JSR;
				wd := Jsrop
				END
			ELSE 		{Ch <> 'R'}
				Error := unkopname
			END
		END {case of Ch}
		END
	ELSE 				{Ch not in ['E','G',...]}
		Error := unkopname
	END; {procedure getJop}


	PROCEDURE getLop (VAR op: opkind; VAR wd: word);
    {Get an operator or directive name that begins with 'L'}
	BEGIN
	getch;
	IF Ch IN ['A', 'D'] THEN
		CASE Ch OF
		'A':  BEGIN
			getch;
			IF Ch = 'B' THEN BEGIN
				getch;
				IF Ch = 'E' THEN BEGIN
					getch;
					IF Ch = 'L' THEN
						op := LABELDIR
					ELSE
						Error := unkopname
					END
				ELSE
					Error := unkopname
				END
			ELSE
				Error := unkopname
			END;
		'D':  BEGIN
			getch;
			IF Ch = 'A' THEN BEGIN
				op := LDA;
				wd := Ldaop
				END
			ELSE BEGIN
				op := LD;
				wd := Ldop;
				Saved := true
				END
			END
		END {case Ch of}
	ELSE
		Error := unkopname
	END; {procedure getLop}


	PROCEDURE getRop (VAR op: opkind; VAR wd: word);
    {Get an operator or directive name that begins with 'R'}
	BEGIN
	getch;
	IF Ch = 'D' THEN BEGIN
		op := RD;
		getch;
		IF Ch IN ['B', 'C', 'F', 'H', 'I', 'N', 'O', 'S'] THEN
			CASE Ch OF
			'B':  BEGIN
				getch;
				IF Ch = 'D' THEN
					wd := Rdop + [8]
				ELSE IF Ch = 'W' THEN
					wd := Rdop + [7, 8]
				ELSE
					Error := unkopname
				END;
			'C':  BEGIN
				getch;
				IF Ch = 'H' THEN
					wd := Rdop + [10]
				ELSE
					Error := unkopname
				END;
			'F': 
				wd := Rdop + [7];
			'H':  BEGIN
				getch;
				IF Ch = 'D' THEN
					wd := Rdop + [8, 9]
				ELSE IF Ch = 'W' THEN
					wd := Rdop + [7, 8, 9]
				ELSE
					Error := unkopname
				END;
			'I': 
				wd := Rdop;
			'N':  BEGIN
				getch;
				IF Ch = 'L' THEN
					wd := Rdop + [7, 8, 10]
				ELSE
					Error := unkopname
				END;
			'O':  BEGIN
				getch;
				IF Ch = 'D' THEN
					wd := Rdop + [9]
				ELSE IF Ch = 'W' THEN
					wd := Rdop + [7, 9]
				ELSE
					Error := unkopname
				END;
			'S':  BEGIN
				getch;
				IF Ch = 'T' THEN
					wd := Rdop + [7, 10]
				ELSE
					Error := unkopname
				END
			END {case of Ch}
		ELSE 				{Ch not in ['B','C',...]}
			Error := unkopname
		END
	ELSE IF Ch = 'E' THEN BEGIN
		getch;
		IF Ch = 'A' THEN BEGIN
			getch;
			IF Ch = 'L' THEN
				op := REALDIR
			ELSE
				Error := unkopname
			END
		ELSE 				{Ch <> 'A'}
			Error := unkopname
		END
	ELSE 				{Ch <> 'E'}
		Error := unkopname
	END; {procedure getRop}


	PROCEDURE getSop (VAR op: opkind; VAR wd: word);
    {Get an operator or directive name that begins with 'S'}
	BEGIN
	getch;
	IF Ch IN ['K', 'L', 'R', 'T'] THEN
		CASE Ch OF
		'K':  BEGIN
			getch;
			IF Ch = 'I' THEN BEGIN
				getch;
				IF Ch = 'P' THEN
					op := SKIPDIR
				ELSE
					Error := unkopname
				END
			ELSE 		{Ch <> 'I'}
				Error := unkopname
			END;
		'L':  BEGIN
			getch;
			IF Ch IN ['C', 'E', 'O', 'Z'] THEN BEGIN
				op := SL;
				CASE Ch OF
				'C':  BEGIN
					getch;
					IF Ch IN ['C', 'E', 'O', 'Z'] THEN
						CASE Ch OF
						'C': 
							wd := Slop + [4, 5, 6];
						'E': 
							wd := Slop + [5, 6];
						'O': 
							wd := Slop + [4, 6];
						'Z': 
							wd := Slop + [6]
						END {case of Ch}
					ELSE BEGIN
						Saved := true;
						wd := Slop + [4, 5]
						END {else}
					END;
				'E': 
					wd := Slop + [5];
				'O': 
					wd := Slop + [4];
				'Z': 
					wd := Slop
				END {case Ch of}
				END
			ELSE
				Error := unkopname
			END;
		'R':  BEGIN
			getch;
			IF Ch IN ['C', 'E', 'O', 'Z'] THEN BEGIN
				op := SR;
				CASE Ch OF
				'C':  BEGIN
					getch;
					IF Ch IN ['C', 'E', 'O', 'Z'] THEN
						CASE Ch OF
						'C': 
							wd := Srop + [4, 5, 6];
						'E': 
							wd := Srop + [5, 6];
						'O': 
							wd := Srop + [4, 6];
						'Z': 
							wd := Srop + [6]
						END {case of Ch}
					ELSE BEGIN
						Saved := true;
						wd := Srop + [4, 5]
						END {else}
					END;
				'E': 
					wd := Srop + [5];
				'O': 
					wd := Srop + [4];
				'Z': 
					wd := Srop
				END {case Ch of}
				END
			ELSE
				Error := unkopname
			END;
		'T':  BEGIN
			getch;
			IF Ch = 'O' THEN BEGIN
				op := STO;
				wd := Stoop
				END
			ELSE IF Ch = 'R' THEN BEGIN
				getch;
				IF Ch = 'I' THEN BEGIN
					getch;
					IF Ch = 'N' THEN BEGIN
						getch;
						IF Ch = 'G' THEN
							op := STRINGDIR
						ELSE
							Error := unkopname
						END
					ELSE 	{Ch <> 'N'}
						Error := unkopname
					END
				ELSE 		{Ch <> 'I'}
					Error := unkopname
				END
			ELSE 		{Ch <> 'R'}
				Error := unkopname
			END
		END {case Ch of}
	ELSE
		Error := unkopname
	END; {procedure getSop}


	PROCEDURE getWop (VAR op: opkind; VAR wd: word);
    {Get an operator or directive name that begins with 'W'}
	BEGIN
	getch;
	IF Ch = 'R' THEN BEGIN
		op := WR;
		getch;
		IF Ch IN ['B', 'C', 'F', 'H', 'I', 'N', 'O', 'S'] THEN
			CASE Ch OF
			'B':  BEGIN
				getch;
				IF Ch = 'D' THEN
					wd := Wrop + [8]
				ELSE IF Ch = 'W' THEN
					wd := Wrop + [7, 8]
				ELSE
					Error := unkopname
				END;
			'C':  BEGIN
				getch;
				IF Ch = 'H' THEN
					wd := Wrop + [10]
				ELSE
					Error := unkopname
				END;
			'F': 
				wd := Wrop + [7];
			'H':  BEGIN
				getch;
				IF Ch = 'D' THEN
					wd := Wrop + [8, 9]
				ELSE IF Ch = 'W' THEN
					wd := Wrop + [7, 8, 9]
				ELSE
					Error := unkopname
				END;
			'I': 
				wd := Wrop;
			'N':  BEGIN
				getch;
				IF Ch = 'L' THEN
					wd := Wrop + [7, 8, 10]
				ELSE
					Error := unkopname
				END;
			'O':  BEGIN
				getch;
				IF Ch = 'D' THEN
					wd := Wrop + [9]
				ELSE IF Ch = 'W' THEN
					wd := Wrop + [7, 9]
				ELSE
					Error := unkopname
				END;
			'S':  BEGIN
				getch;
				IF Ch = 'T' THEN
					wd := Wrop + [7, 10]
				ELSE
					Error := unkopname
				END
			END {case of Ch}
		ELSE
			Error := unkopname
		END
	ELSE 				{Ch <> 'R'}
		Error := unkopname
	END; {procedure getWop}


	PROCEDURE proline;
    {process the current line of input}
		VAR
			wd, wd2: word;
			b1, b2: bytee;
			op: opkind;
			reg: integer;
			twowds: boolean;
			i1, i2: integer;
			id: name;
			idrec: symptr;
	BEGIN
	twowds := false;
	Error := notErr;
	IF Ch IN ['B', 'C', 'F', 'H', 'I', 'J', 'L', 'N', 'R', 'S', 'T', 'W'] THEN
		CASE Ch OF
		'B': 
			getBop(op, wd);
		'C':  BEGIN
			getch;
			IF Ch = 'L' THEN BEGIN
				getch;
				IF Ch = 'R' THEN BEGIN
					op := CLR;
					wd := Srop
					END
				ELSE 		{Ch <> 'R'}
					Error := unkopname
				END
			ELSE 		{Ch <> 'L'}
				Error := unkopname
			END;
		'F': 
			getFop(op, wd);
		'H':  BEGIN
			getch;
			IF Ch = 'A' THEN BEGIN
				getch;
				IF Ch = 'L' THEN BEGIN
					getch;
					IF Ch = 'T' THEN BEGIN
						op := HALT;
						wd := Haltop
						END
					ELSE 	{Ch <> 'T'}
						Error := unkopname
					END
				ELSE 		{Ch <> 'L'}
					Error := unkopname
				END
			ELSE 		{Ch <> 'A'}
				Error := unkopname
			END;
		'I': 
			getIop(op, wd);
		'J': 
			getJop(op, wd);
		'L': 
			getLop(op, wd);
		'N':  BEGIN
			getch;
			IF Ch = 'O' THEN BEGIN
				getch;
				IF Ch = 'P' THEN BEGIN
					op := NOP;
					wd := Jop + [7, 8, 9]
					END
				ELSE 		{Ch <> 'P'}
					Error := unkopname
				END
			ELSE 		{Ch <> 'O'}
				Error := unkopname
			END;
		'R': 
			getRop(op, wd);
		'S': 
			getSop(op, wd);
		'T':  BEGIN
			getch;
			IF Ch = 'R' THEN BEGIN
				getch;
				IF Ch = 'N' THEN BEGIN
					getch;
					IF Ch = 'G' THEN BEGIN
						op := TRNG;
						wd := Trngop
						END
					ELSE	{Ch <> 'G'}
						Error := unkopname
					END
				ELSE		{Ch <> 'N'}
					Error := unkopname
				END
			ELSE		{Ch <> 'R'}
				Error := unkopname
			END;
		'W': 
			getWop(op, wd)
		END {case of Ch}
	ELSE 					{Ch not a valid first letter}
		Error := unkopname;
	IF Error = notErr THEN BEGIN
		CASE op OF
		CLR:  BEGIN			{need to find a reg address}
			reg := getregaddr;
			IF Error = notErr THEN BEGIN
				inregaddr(wd, reg, 10);
				insertmem(wd)
				END {if Error = notErr}
			END;			{SL, SR case}
		SL, SR:  BEGIN			{need to find a reg address}
			reg := getregaddr;
			IF Error = notErr THEN
				inregaddr(wd, reg, 10);
			getch;
			IF Ch <> ',' THEN BEGIN
				warn(missingcomma);
				Saved := true
				END;
			scanint(Icr.wdf);
			IF Error = notErr THEN BEGIN
				reg := Icr.sintf;
				IF reg = 16 THEN
					reg := 0;
				IF (reg < 16) AND (reg >= 0) THEN BEGIN
					inregaddr(wd, reg, 3);
					insertmem(wd)
					END
				ELSE
					Error := badshftamt
				END {if Error = notErr}
			END;			{SL, SR case}
		OIN, IA, IS, IM, IDENT, FN, FA, FS, FM, FD, BI, BO, BA, IC, FC, JSR, BKT, LD, STO, LDA, FLT, FIX, TRNG:  BEGIN
			reg := getregaddr;
			IF Error = notErr THEN BEGIN
				inregaddr(wd, reg, 10);
				getch;
				IF Ch <> ',' THEN BEGIN
					warn(missingcomma);
					Saved := true
					END; {if Ch <> ','}

				getgenaddr(op, wd, wd2, twowds);

				IF Error = notErr THEN
					IF twowds THEN BEGIN
						insertmem(wd);
						insertmem(wd2)
						END
					ELSE
						insertmem(wd)
				END {if Error = notErr}
			END;			{two-address case}
		J:  BEGIN
			getgenaddr(op, wd, wd2, twowds);
			IF Error = notErr THEN
				IF twowds THEN BEGIN
					insertmem(wd);
					insertmem(wd2)
					END
				ELSE
					insertmem(wd)
			END;			{one general address case}
		RD, WR:  BEGIN
			twowds := false;
			IF NOT ((10 IN wd) AND (8 IN wd)) THEN
				getgenaddr(op, wd, wd2, twowds);
			IF Error = notErr THEN
				IF twowds THEN BEGIN
					insertmem(wd);
					insertmem(wd2)
					END
				ELSE
					insertmem(wd)
			END;			{one general address case}
		NOP, HALT: 
			insertmem(wd);
		INTDIR:  BEGIN
			scanint(wd);
			wordtobytes(wd, Mem[Lc], Mem[Lc + 1]);
			Lc := Lc + 2
			END;
		REALDIR:  BEGIN
			scanreal(wd, wd2);
			wordtobytes(wd, Mem[Lc], Mem[Lc + 1]);
			wordtobytes(wd2, Mem[Lc + 2], Mem[Lc + 3]);
			Lc := Lc + 4
			END;
		SKIPDIR:  BEGIN
			scanint(wd);
			Icr.wdf := wd;
			i1 := Icr.sintf;
			IF i1 < 0 THEN
				warn(badskip)
			ELSE
				Lc := Lc + i1
			END;
		STRINGDIR:  BEGIN
			scanstr
			END;
		LABELDIR:  BEGIN
			getch;
			IF Ch IN ['a', 'z', 'A'..'Q', 'S'..'Z', '$'] THEN BEGIN
				scanname(id);
				idrec := findname(id);
				IF idrec = NIL THEN
					idrec := inname(id)
				ELSE BEGIN
					IF idrec^.loc >= 0 THEN
						warn(namedefined);
					i1 := idrec^.patch;
					Icr.sintf := Lc;
					wordtobytes(Icr.wdf, b1, b2);
					WHILE i1 >= 0 DO BEGIN
						bytestoword(Mem[i1], Mem[i1 + 1], Icr.wdf);
						i2 := Icr.sintf;
						Mem[i1] := b1;
						Mem[i1 + 1] := b2;
						i1 := i2
						END {while}
					END; {else}
				idrec^.patch := -1;
				idrec^.loc := Lc
				END
			ELSE
				Error := invalidname
			END
		END {case of op}
		END {if Error = notErr}
	END; {procedure proline}

	PROCEDURE initopcodes;
	BEGIN
	Inop := [];
	Iaop := [11];
	Isop := [12];
	Imop := [11, 12];
	Idop := [13];
	Fnop := [11, 13];
	Faop := [12, 13];
	Fsop := [11, 12, 13];
	Fmop := [14];
	Fdop := [11, 14];
	Biop := [12, 14];
	Boop := [11, 12, 14];
	Baop := [13, 14];
	Icop := [11, 13, 14];
	Fcop := [12, 13, 14];
	Jsrop := [11, 12, 13, 14];
	Bktop := [15];
	Ldop := [11, 15];
	Stoop := [12, 15];
	Ldaop := [11, 12, 15];
	Fltop := [13, 15];
	Fixop := [11, 13, 15];
	Jop := [12, 13, 15];
	Srop := [11, 12, 13, 15];
	Slop := [14, 15];
	Rdop := [11, 14, 15];
	Wrop := [12, 14, 15];
	Trngop := [11, 12, 14, 15];
	Haltop := [11, 12, 13, 14, 15]
	END; {procedure initopcodes}

BEGIN {main}
showDrawing;  {Think Pascal}
moveTo(40, 20);
drawstring('SAM2, the MACC2 Symbolic Assembler');
moveTo(40, 35);
DrawString('Name of the Code File? ');
CodeFileName := oldFileName('Name of the Code File? ');
IF CodeFileName <> '' THEN BEGIN
	Reset(InFile, CodeFileName);
	clearLine(35);
	moveTo(40, 35);
	DrawString('Name of the Listing File? ');
	ListingFileName := NewFileName('Name of the Listing File? ');
	IF ListingFileName = '' THEN
		Rewrite(outFile, 'Listing.T')
	ELSE
		Rewrite(outFile, ListingFileName);
	moveTo(320, 20);
	drawString('WORKING...');

    {assign ( InFile, 'CODEFILE' );}
    {reset ( InFile, 'CODEFILE' );}
	symbols := NIL;
	initopcodes;
	cLine := 1;
	Lc := 0;
	Errs := false;
	Warning := false;
	Saved := false;
	Listing := true;
{InFile^ = '@';}

	IF Listing AND NOT eof(InFile) THEN BEGIN
		writeln(outFile);
		write(outFile, Lc : 4, ': ');
		END;
{listing := false;}
	WHILE NOT eof(InFile) DO BEGIN
		getch;
		WHILE NOT eof(InFile) AND ((Ch = '%') OR (Ch = ' ') OR (Ch = '@')) DO BEGIN
			readln(InFile);
			IF ch = '@' THEN
				listing := NOT listing;
			IF Listing AND NOT eof(InFile) THEN BEGIN (* Dana -- added 'and not eof' *)
				writeln(outFile);
				write(outFile, Lc : 4, ': ');
				END;
			getch;
			cLine := cLine + 1
			END;
		IF NOT eof(InFile) THEN BEGIN
			Error := notErr;
			Morewarn := false;
			proline;
			getch;
			IF (Ch <> '%') AND (Ch <> ' ') THEN BEGIN
				warn(textfollows);
				WHILE NOT eoln(InFile) DO BEGIN
					read(InFile, Ch);
					IF Listing THEN
						write(outFile, Ch)
					END
				END; {while}
			readln(InFile);
			IF Listing AND (NOT eof(InFile)) THEN BEGIN (* Dana -- added 'and not eof' *)
				writeln(outFile);
				write(outFile, Lc : 4, ': ');
				END;
			IF Error <> notErr THEN BEGIN
				Errs := true;
				write(outFile, 'ERROR -- ');
				CASE Error OF
				invalidname: 
					write(outFile, 'Invalid Name in Label Directive');
				unkopname: 
					write(outFile, 'Unknown Operation or Directive Name');
				badstring: 
					write(outFile, 'Improper String Directive -- Missing "');
				badgenaddr: 
					write(outFile, 'Improperly Formed General Address');
				badregaddr: 
					write(outFile, 'Register Address Out of Range');
				badint: 
					write(outFile, 'Improperly Formed Integer Constant');
				badreal: 
					write(outFile, 'Improperly Formed Real Constant');
				badstr: 
					write(outFile, 'Improperly Formed String Constant');
				badname: 
					write(outFile, 'Improperly Formed Name');
				illregaddr: 
					write(outFile, 'Direct Register Address not Permitted');
				illimedaddr: 
					write(outFile, 'Immediate Address not Permitted');
				badshftamt: 
					write(outFile, 'Shift Amount not in Range')
				END; {case Error of}
				writeln(outFile, ' detected on line ', cLine : 1)
				END;
			WHILE Morewarn DO
				printwarn;
			cLine := cLine + 1
			END
		END; {while}

	checktab(Symbols);

    {assign ( memf, 'OBJ' );}
	rewrite(memf, 'OBJ'); (* write an empty file if necessary *)
	IF NOT Errs AND NOT Warning THEN BEGIN
		FOR i := 0 TO Lc - 1 DO BEGIN
			Bcr.bytef := Mem[i];
			write(memf, Bcr.hintf)
			END; {for}
		END;

	close(InFile);
	close(memf);
	close(OutFile);
	noError := GetFInfo(ListingFileName, vRefNum, finderInfo);
	finderInfo.fdType := 'TEXT';
	finderInfo.fdCreator := 'QED1';
	noError := SetFInfo(ListingFileName, vRefNum, finderInfo);
	END;
END.